/*
 * Latency Tester: Use the TSC to measure variation in code execution caused by
 * hardware or the operating system
 *
 * (C) 2009 The Linux Foundation
 *      Christoph Lameter <cl@linux-foundation.org>
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <getopt.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <asm/unistd.h>

#ifdef __IGNORE_getcpu
/*
 * getcpu is a vsyscall.
 * This means that sched_getcpu is supported through sched.h
 */
#include <sched.h>

#else

/* No support for getcpu() in glibc */
#define __NR_getcpu 318
/*
 * Fallback implementation if the system does not provide
 * sched_getcpu() (f.e. RH4)
 */
int sched_getcpu(void)
{
	unsigned cpu = 0;
	unsigned node;
	syscall(__NR_getcpu, &cpu, &node, NULL);
	return cpu;
}
#endif

/*
 * From Linux 2.6.29 x86 arch code:
 *
 * both i386 and x86_64 returns 64-bit value in edx:eax, but gcc's "A"
 * constraint has different meanings. For i386, "A" means exactly
 * edx:eax, while for x86_64 it doesn't mean rdx:rax or edx:eax. Instead,
 * it means rax *or* rdx.
 */
#ifdef __LP64__
#define DECLARE_ARGS(val, low, high)    unsigned low, high
#define EAX_EDX_VAL(val, low, high)     ((low) | ((unsigned long long)(high) << 32))
#define EAX_EDX_ARGS(val, low, high)    "a" (low), "d" (high)
#define EAX_EDX_RET(val, low, high)     "=a" (low), "=d" (high)
#else
#define DECLARE_ARGS(val, low, high)    unsigned long long val
#define EAX_EDX_VAL(val, low, high)     (val)
#define EAX_EDX_ARGS(val, low, high)    "A" (val)
#define EAX_EDX_RET(val, low, high)     "=A" (val)
#endif

typedef unsigned long long cycle_t;

inline cycle_t get_cycles() {
        DECLARE_ARGS(val, low, high);

        asm volatile("rdtsc" : EAX_EDX_RET(val, low, high));
        return EAX_EDX_VAL(val, low, high);
}

#define CYC2FS_SCALE_FACTOR 0
#define CYC2NS_SCALE_FACTOR 10 /* 2^10, carefully chosen */
#define CYC2US_SCALE_FACTOR 20 /* 2^20 */

#define FSEC_PER_MSEC 1000000000L
#define NSEC_PER_MSEC 1000000L
#define USEC_PER_MSEC 1000L

static int cyc2fs;
static int cyc2ns;
static int cyc2us;

static inline unsigned long long cycles_2_fs(cycle_t cyc)
{
	return (cyc * cyc2fs) >> CYC2FS_SCALE_FACTOR;
}

static inline unsigned long long cycles_2_ns(cycle_t cyc)
{
	return (cyc * cyc2ns) >> CYC2NS_SCALE_FACTOR;
}

static inline unsigned long long cycles_2_us(cycle_t cyc)
{
	return (cyc * cyc2us) >> CYC2US_SCALE_FACTOR;
}

static inline void set_cyc2ns_scale(unsigned long cpu_khz)
{
	cyc2fs = (FSEC_PER_MSEC << CYC2FS_SCALE_FACTOR)/cpu_khz;
	cyc2ns = (NSEC_PER_MSEC << CYC2NS_SCALE_FACTOR)/cpu_khz;
	cyc2us = (USEC_PER_MSEC << CYC2US_SCALE_FACTOR)/cpu_khz;
}

unsigned long long timestamp_nanoseconds(void)
{	cycle_t x = get_cycles();

	return cycles_2_ns(x);
}

unsigned long long timestamp_femtoseconds(void)
{
	cycle_t x = get_cycles();

	return cycles_2_fs(x);
}

unsigned long long timestamp_microseconds(void)
{
	cycle_t x = get_cycles();

	return cycles_2_us(x);
}

/* REP NOP (PAUSE) is a good thing to insert into busy-wait loops. */
void cpu_relax(void)
{
        asm volatile("rep; nop" ::: "memory");
}

/* Hardware description */
unsigned int processors;
unsigned int cores;
unsigned long cacheline_size;
unsigned long cpu_l3_cache_size;
unsigned long cpu_l2_cache_size;
unsigned long cpu_l1d_cache_size;
unsigned int cpu_l1i_cache_size;
char *cpu_name;
unsigned int cpu_family;
unsigned int cpu_model;
unsigned int cpu_stepping;
unsigned long long cpu_clock_frequency;

int getline(char *buffer, int size, FILE *f)
{
	char *x;

	x = fgets(buffer, size, f);

	if (!x)
		return 0;

	buffer[strlen(buffer) - 1] = 0;
	return 1;
}

/* Initialize library */
void ll_init(int verbose)
{
	char buffer[200];
	char * tag = buffer;
	char *text;
	FILE *f;
	unsigned int number;

	f = fopen("/proc/cpuinfo","r");
	if (!f) {
		fprintf(stderr,"Cannot read /proc/cpuinfo\n");
		abort();
	}
	while (getline(buffer,sizeof(buffer), f)) {
		/* find the : */
		char *p = index(buffer, ':');

		if (p) {
			char *q = p - 1;

			text = p + 2;
			while (q >= buffer && (*q == ' ' || *q == 9))
				q--;

			q[1] = 0;

			number = atoi(text);
			if (strcmp(tag, "processor")==0) {
				if (number > processors)
					processors = number;
			} else if (strcmp(tag, "core id")==0) {
				if (number > cores)
					cores = number;
			} else if (strcmp(tag, "cpu family")==0) {
				cpu_family = number;
			} else if (strcmp(tag, "model")==0) {
				cpu_model = number;
			} else if (strcmp(tag, "model name")==0) {
				cpu_name = strdup(text);;
			} else if (strcmp(tag, "stepping")==0) {
				cpu_stepping = number;
			} else if (strcmp(tag, "cpu MHz")==0) {
				cpu_clock_frequency = atof(text) * 1000000.0;
			} else if (strcmp(tag, "cache size")==0) {
				cpu_l3_cache_size = number * 1000;
			} else if (strcmp(tag, "cache_alignment")==0) {
				cacheline_size = number;
			}
		}
	}
	fclose(f);
	if (!cpu_clock_frequency) {
		fprintf(stderr,"Unable to determine processor clock\n");
		abort();
	}
	set_cyc2ns_scale(cpu_clock_frequency / 1000.0);

	cores++;
	processors++;

	if (verbose)
		printf("CPUs: Freq=%.2fGhz Processors=%u Cores=%u cacheline_size=%lu %s\n",
			cpu_clock_frequency / 1000000000.0, processors, cores, cacheline_size, cpu_name);
}

