/*
 * Multicast Testing code
 *
 * (C) 2009 Linux Foundation
 *		Christoph Lameter <cl@linux-foundation.org>
 */
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdarg.h>
#include <getopt.h>
#include <pthread.h>
#include <linux/types.h>
#include <math.h>
#include <errno.h>

#define MAGIC 0xFFEEDDCC

enum control_message { NOOP, TIME_SYNC_REQUEST, TIME_SYNC_FINAL, TIME_SYNC_REPLY, END_OF_TEST, END_OF_PROGRAM };

char *control_text[] = {
	"No Operation",
	"Time sync request",
	"Time sync final",
	"Time sync reply",
	"End of test"
};

struct msg {
	int magic;
	int ad;
	int port;
	__u64 timestamp;
	int burst;
	int seqid;
	int length;
	char payload[10];
};

void fatal(const char *x, ...)
{
	va_list ap;

	va_start(ap, x);
	vfprintf(stderr, x, ap);
	va_end(ap);
	exit(EXIT_FAILURE);
}

static inline long minimum(int a, int b)
{
	if (a < b)
		return a;
	return b;
}

static inline __u64 get_timestamp(void)
{
	struct timespec spec;

	clock_gettime(CLOCK_REALTIME, &spec);
	return spec.tv_nsec + spec.tv_sec * 1000000000LL;
}

int port = 9000;	/* Port to listen to or send to */
int samples = 10;	/* Number of samples to take */
int runtime = 30;	/* Maximum runtime */
int ttl = 20;		/* Set ttl */
int length = 300;
int histogram = 0;
int time_sync = 0;
int time_offset = 0;
int latencies = 0;	/* Display latencies */
int aggregate = 0;	/* Make OS coalesce packets */
int terminate_remote = 0; /* Terminate remote listener after test */
int warmup = 0;
int class = 3;		/* 1 = A 2 = B 3 = C */

int origport = 9001;	/* Origination port for ping */
struct sockaddr_in bindaddr;	/* IP address to bind to */
struct sockaddr_in addr;

#define MAX_MULTICASTGROUPS 10
int multicastgroups;
struct sockaddr_in multicastgroup[MAX_MULTICASTGROUPS];


int verbose = 0;
int rate = 0;		/* Limit rate */
int burstsize = 1;
struct timespec wait_time = { 0, 0 };		/* Pause between packets */

int inblocksize = 10000;	/* Size of the block to send */
int outblocksize = 1000;	/* Size of the block to send */

int yes = 1;

int sock;

int open_sock(struct sockaddr_in *sa)
{
	int err;
	int i;
	struct sockaddr inad;

	sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sock < 0) {
		perror("socket");
		return -2;
	}

	err = setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes));
	if (err) {
		perror("reuseaddr");
		return -3;
	}

	((struct sockaddr_in *)&inad)->sin_family = AF_INET;
	((struct sockaddr_in *)&inad)->sin_addr.s_addr = bindaddr.sin_addr.s_addr;
	((struct sockaddr_in *)&inad)->sin_port = bindaddr.sin_port;
	if (bind(sock, &inad, sizeof(inad))) {
		perror("bind");
		return -4;
       	}

	if (ttl >= 0) {
		err = setsockopt(sock, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl));
		if (err) {
			perror("setsockopt IP_MULTICAST_TTL\n");
			return -5;
		}
	}

	/* Subscribe to multicast groups */
	for (i = 0; i < multicastgroups; i++) {
		struct ip_mreq mreq;

		mreq.imr_multiaddr.s_addr = multicastgroup[i].sin_addr.s_addr;
		mreq.imr_interface.s_addr = htonl(INADDR_ANY);
		if (setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0) {
			perror("setsocktopt");
			return -6;
		} else if (verbose)
			printf("Joined mcast group %s\n", strdupa(inet_ntoa(mreq.imr_multiaddr)));
	}
	return sock;
}

struct sockaddr_in sendaddr;
unsigned sendaddrlen;

char msg[10000];

#define MAX_US 1000000
int hist[MAX_US];

void mcast_time_sync_msg(struct sockaddr_in *, int, struct sockaddr *, int);

int mcast_receive(struct sockaddr_in *sa)
{
	int err;
	int nbytes;
	int seqid;
	int burst;
	int messages;
	struct sockaddr inad;
	unsigned inadlen = sizeof(inad);
	struct msg *inmsg;
	long diff;
	unsigned long min, max;
	unsigned long long sum, sqsum;
	double avg, run_seconds;
	int outliers, sequence_error;
	long correction_factor;
	int corrupted;
	unsigned long long sync_time, start_time, run_time;
	unsigned long bytes;
	unsigned long kbytes;

redo:
	sequence_error = 0;
	seqid = 1;
	messages = 0;
	sum = 0;
	sqsum = 0;
	min = 0;
	max = 0;
	outliers = 0;
	corrupted = 0;
	correction_factor = 0;
	bytes = 0;
	memset(hist, 0, sizeof(hist));

	inmsg = (void *)msg;
	err = open_sock(sa);

	if (err < 0)
		fatal("Failed to open socket code =%d\n", err);
	if (sock < 0)
		return err;

	printf("Waiting for udp traffic on port %d ...\n", ntohs(sa->sin_port));

redo2:
	sendaddrlen = sizeof(sendaddr);
	nbytes = recvfrom(sock, inmsg, sizeof(msg), MSG_PEEK,
				(struct sockaddr *)&sendaddr, &sendaddrlen);
	if (nbytes < 0) {
		perror("recfrom");
                printf("Receive failed. %d\n", nbytes);
                return 6;
        }

	if (inmsg->burst <= 0) {
		if (inmsg->burst == -END_OF_PROGRAM) {
			goto out;
		}
		if (inmsg->burst == -TIME_SYNC_REQUEST) {
			sync_time = get_timestamp();
			mcast_time_sync_msg(sa, -2, (struct sockaddr *)&sendaddr, sendaddrlen);
		} else if (inmsg->burst == -TIME_SYNC_FINAL) {
			unsigned long long sync2_time, other_end_time, this_end_time;
			long new_factor;

			sync2_time = get_timestamp();

			/* A reply to the time sync */
			other_end_time = inmsg->timestamp;

			/* We expect communication to be symmetric. The other end will therefore
			 * reply to the message at half time between sync and syn2 times.
			 * The difference is the necessary correction factor
			 */
			this_end_time = sync_time + (sync2_time - sync_time) / 2;
			new_factor = other_end_time - this_end_time;

			if (abs(new_factor) < abs(correction_factor) || !correction_factor)
				correction_factor = new_factor;

		} else if (verbose)
			printf("** Ignoring spurious control message: %s\n", control_text[-inmsg->burst]);

		recvfrom(sock, inmsg, sizeof(msg), 0, (struct sockaddr *)&sendaddr, &sendaddrlen);
		goto redo2;
	}

	if (correction_factor)
		printf("Estimated time correction factor = %ld ns\n", correction_factor);


	printf("Msg Peek resulted in receiving %d bytes. Expecting %d messages\n", nbytes, inmsg->burst);
	burst = inmsg->burst;

	start_time = get_timestamp();
	nbytes = 0;
	do {
		if (nbytes > inmsg->length) {
			if (verbose)
				printf(" ");
			nbytes -= inmsg->length;
			memcpy(inmsg, (char *)inmsg + inmsg->length, nbytes);
		} else
			nbytes = recvfrom(sock, inmsg, sizeof(msg),
								0, &inad, &inadlen);
		if (nbytes < 0) {
			perror("recvfrom");
			break;
		}
		if (inmsg->magic != MAGIC) {
			printf("*** Corrupted packet or packet from another source received. Ignored\n");
			corrupted++;
			continue;
		}
		if (inmsg->burst != burst)
			break;

		diff = get_timestamp() - inmsg->timestamp + correction_factor;
		if (inmsg->seqid != seqid) {
			printf("*** Message out of sequence. Expected %d received %d length=%d latency=%7.2fusecs\n",
				seqid, inmsg->seqid, inmsg->length, diff / 1000.0);
			seqid = inmsg->seqid;
			sequence_error++;
		}
		if (verbose)
			printf("%d/%d Received %d bytes from %s. length=%d Latency=%7.2fusec\n",
				inmsg->seqid, inmsg->burst, nbytes,
				strdupa(inet_ntoa(((struct sockaddr_in *)&inad)->sin_addr)),
				inmsg->length,
				diff / 1000.0);

		if (latencies) {
			if (diff < 0 || diff > MAX_US * 1000) {
				if (outliers < 10) {
					printf("%d/%d Outlier #%d: Latency=%7.2fusecs\n",
						inmsg->seqid, inmsg->burst, outliers + 1, diff / 1000.0);
					if (outliers == 9)
						printf("Too many outliers. Stopped messages\n");
				}
				outliers++;
			} else {
				int udiff;

				if (diff < min || min == 0)
					min = diff;
				if (diff > max)
					max = diff;
				udiff = diff / 1000;
				hist[ udiff] ++;
				sum += udiff;
				sqsum += udiff * udiff;
			}
		}
		messages++;
		seqid++;
		bytes += nbytes;

	} while (burst >= seqid);
	run_time = get_timestamp() - start_time;
	close(sock);

	run_seconds = (double)run_time / (double)1000000000.0;
	kbytes = (bytes + 500) / 1000;

	if (seqid <= burst)
		printf("*** Messages from %d to %d lost\n", seqid - 1, burst);

	printf("\n----------------Summary-------------------------------------------\n");
	avg = (double) sum / messages;
	printf("Messages : %d of %d              %.0f msgs/sec\n",
			messages, burst,
			messages / run_seconds);
	if (latencies && outliers < 10) {
		printf("Mininum  : %7.2fusec                %.0f kbytes/sec\n",
			min / 1000.0, kbytes / run_seconds);
		printf("Maximum  : %7.2fusec                %lu kbytes\n",
			max / 1000.0, kbytes);
		printf("Average  : %7.2fusec                %.1f seconds\n",
			avg, run_seconds);
		printf("Stddev   : %7.2fusec\n", sqrt((double)sqsum / messages - avg * avg));
	} else if (latencies)
		printf("Latency statistics not available. Too many outliers.\n");

	if (burst > messages)
		printf("Lost     : %d\n", burst - messages);
	if (outliers)
		printf("Outliers : %d\n", outliers);
	if (corrupted)
		printf("Corrupted: %d\n", corrupted);
	if (sequence_error)
		printf("SeqErrors: %d\n", sequence_error);

	if (latencies && histogram) {
		int i, count, scale, max;

		printf("\nMicrosec\tPercent\tSamples\n");


		/* Conflate array to have less than 20 entries */
		count = 999999;
		scale = 1;
		max = MAX_US;

		count = 0;
		for (i = 0; i < max; i++)
			if (hist[i])
				count++;

		while (count > 20) {
			int newmax = max / 2;

			/* Conflate array */
			for (i = 0; i < newmax; i++)
				hist[i] = hist[ 2* i ]  + hist [ 2*i + 1 ];

			if (max % 2)
				hist[newmax - 1] += hist[max - 1];

			max = newmax;
			scale = scale * 2;

			/* Recount */
			count = 0;
			for (i = 0; i < max; i++)
				if (hist[i])
					count++;
		}

		for(i=0; i < max; i++)
			if (hist[i])
				printf("%4.0f-%.0f\t%4.0f%%\t%8u\n",
					(double)i * scale,
					(double)(i + 1) * scale,
					hist[i] * (double)100.0 / messages,
					hist[i]);
	}
	printf("-----------------------------------------------------------------\n");

	goto redo;
out:
	return 0;
}

struct msg outmsg;

int mcast_send(struct sockaddr_in *sa)
{
	int i;
	int nbytes;

	open_sock(sa);

	if (time_sync) {
		struct sockaddr_in syncaddr;
		unsigned syncaddrlen;

		for (i = 0; i < 10; i++) {
			if (verbose)
				printf("Sending time sync request\n");
			/* Multicastd request for a global time sync */
			mcast_time_sync_msg(sa, -TIME_SYNC_REQUEST, (struct sockaddr *)&addr, sizeof(addr));
			if (verbose)
				printf("Waiting for time sync response\n");

			/* Unicast response */
			nbytes = recvfrom(sock, &outmsg, sizeof(msg), 0,
					(struct sockaddr *)&syncaddr, &syncaddrlen);
			if (nbytes < 0) {
				perror("recfrom");
       		     		printf("UDP MCAST receive failed %d\n", nbytes);
       				return 6;
		        }

			if (verbose)
				printf("Sending time sync confirmation\n");

			/* Unicast reply */
			mcast_time_sync_msg(sa, -TIME_SYNC_FINAL, (struct sockaddr *)&syncaddr, sizeof(syncaddr));
		}
	}

	printf("udp mcast %s:%d -> %s:%d %d samples ....\n",
				 strdupa(inet_ntoa(bindaddr.sin_addr)), origport,
				 strdupa(inet_ntoa(sa->sin_addr)), port,
				 samples);
	outmsg.magic = MAGIC;
	outmsg.ad = sa->sin_addr.s_addr;
	outmsg.port = htons(origport);
	outmsg.burst = samples;
	outmsg.length = length;

	for (i = 0; i < sizeof(outmsg.payload); i++)
		outmsg.payload[i] = i;

	for (i = 0; i < samples; i++) {

		if (rate && (i % burstsize) == 0) {
			struct timespec tmp;

			wait_time.tv_nsec = 1000000000 / rate * burstsize;
			nanosleep(&wait_time, &tmp);
		}

		outmsg.timestamp = get_timestamp() - time_offset;
		outmsg.seqid = i + 1;
		if (sendto(sock, &outmsg, length,
				(aggregate && burstsize > 1 && (outmsg.seqid % burstsize)) ? MSG_MORE : 0,
				&addr, sizeof(addr)) < 0) {
			perror("sendto");
			break;
		}

	}
	printf("Send %d messages\n", samples);
	mcast_time_sync_msg(sa, -END_OF_TEST, (struct sockaddr *)&addr, sizeof(addr));
	sleep(1);
	mcast_time_sync_msg(sa, terminate_remote ? -END_OF_PROGRAM : -END_OF_TEST,
				 (struct sockaddr *)&addr, sizeof(addr));
	return 0;
}

void mcast_time_sync_msg(struct sockaddr_in *sa, int mode, struct sockaddr *addr, int addrlen)
{
	int i,e;

	outmsg.magic = MAGIC;
	outmsg.ad = sa->sin_addr.s_addr;
	outmsg.port = htons(origport);
	outmsg.burst = mode;

	for (i = 0; i < sizeof(outmsg.payload); i++)
		outmsg.payload[i] = i;

	outmsg.timestamp = get_timestamp() - time_offset;
	outmsg.seqid = 0;

	do {
		if (sendto(sock, &outmsg, sizeof(outmsg), 0, addr, addrlen) < 0) {
			e = errno;
			if (e == ENOBUFS)
				sleep(1);
			else
				perror("sendto");
		}
	} while (e == ENOBUFS);
}

int mcast_warmup(struct sockaddr_in *sa)
{
	int i;
	int nr;

	open_sock(sa);

	switch(class) {
		case 1: nr = 256*256 * 256; break;
		case 2: nr = 256*256;break;
		case 3: nr = 256; break;
	}

	for (i = 0; i < nr; i++) {

		memcpy(&sendaddr, sa, sizeof(sendaddr));
		sendaddr.sin_addr.s_addr = htonl(ntohl(sendaddr.sin_addr.s_addr) + i);
		sendaddr.sin_port = htons(origport);

		if (verbose)
			printf("Sending to %s:%d\n", strdupa(inet_ntoa(sendaddr.sin_addr)), ntohs(sendaddr.sin_port));
		mcast_time_sync_msg(sa, -END_OF_TEST, (struct sockaddr *)&sendaddr, sizeof(sendaddr));
	}
	return 0;
}

void usage(void)
{
	printf("mcasttest 6/4/2009 (C) 2009 Linux Foundation\n\n"
		"mcasttest [destination]\n"
		"-j<mcastgroup>|--join <mcgrp>     Join Multicast group (can be specified multiple times)\n"
		"-p<port>|--port <port>            Change the port to use [9000]\n"
		"-n<samples>|--number <samples>    Specify the number of samples [10]\n"
		"-t<secs>|--time <secs>            Maximum run time of ping[30]\n"
		"-l<length>|--length <length>      Set length of packet\n"
		"-g<port>|--origport <port>        Set origination port [9001]\n"
		"-b<ip>|--bind <ip>                Bind to a certain IP address\n"
		"-h|--help                         Print usage information\n"
		"-d|--details                      Print histogram\n"
		"-s|--sync                         Synchronize time\n"
		"-S<offset>/--sync-offset <off>    Manually specify offset on send\n"
		"-i<bytes>|--inblocksize <bytes>   Maximum number of bytes to accept with one syscall\n"
		"-o<bytes>|--outblocksize <bytes>  Maximum number of bytes to send with one syscall\n"
		"-r<packets>|--rate <packets>      Packets to send per second.\n"
		"-B>packets>|--burst <packets>	   Batchsize for packets (default 1)\n"
		"-L|--latencies                    Report on message latencies (beware of the clock skew!)\n"
		"-X|--exit                         Terminate receive on the other side\n"
		"-A|--aggregate                    Burst mode uses MSG_MORE\n"
		"-W|--warmup	  	           Warm up Multicast range\n"
		"-C|--class A/B/C		   Class of warm up (default is C)\n"
	);
}

struct option opts[] = {
	{ "port", 1, NULL, 'p' },
	{ "number", 1, NULL, 'n' },
	{ "time", 1, NULL, 't' },
	{ "length", 1, NULL, 'l' },
	{ "origport", 1, NULL, 'g' },
	{ "bind", 1, NULL, 'b' },
	{ "help", 0, NULL, 'h' },
	{ "verbose", 0, NULL, 'v' },
	{ "inblocksize", 0, NULL, 'i' },
	{ "outblocksize", 0, NULL, 'o' },
	{ "rate", 1, NULL, 'r' },
	{ "join", 1, NULL, 'j' },
	{ "details", 1, NULL, 'd' },
	{ "sync", 0, NULL, 's' },
	{ "sync-offset", 1, NULL, 'S' },
	{ "burst", 1, NULL, 'B' },
	{ "latencies", 0, NULL, 'L' },
	{ "terminaate", 0, NULL, 'X' },
	{ "aggregate", 0, NULL, 'A' },
	{ "warmup", 1, NULL, 'W' },
	{ "class", 1, NULL, 'C' },
	{ NULL, 0, NULL, 0 }
};

int main(int argc, char *argv[])
{
	int c;

	addr.sin_family = AF_INET;
	bindaddr.sin_family = AF_INET;
	bindaddr.sin_addr.s_addr = htonl(INADDR_ANY);

	while ((c = getopt_long(argc, argv, "AXLvhtedsp:n:t:l:o:b:i:o:r:j:B:S:C:W",
							opts, NULL)) != -1)
	switch (c) {
	case 'h' :
		usage();
		return 0;

	case 'p' :
		port = atoi(optarg);
		if (port < 1024 || port > 60000)
			fatal("Invalid port number %d\n", port);
		break;

	case 'g' :
		origport = atoi(optarg);
		if (origport < 1024 || origport > 60000)
			fatal("Invalid port number %d\n", origport);
		break;

	case 'n' :
		samples = atoi(optarg);
		if (samples < 10 || samples > 1000000000)
			fatal("Invalid number of samples %d\n", samples);
		break;

	case 't' :
		runtime = atoi(optarg);
		if (runtime < 1 || runtime > 360)
			fatal("Invalid duration %d seconds\n", runtime);
		break;

	case 'l' :
		length = atoi(optarg);
		if (length < 40 || length > sizeof(msg))
			fatal("Invalid packet length %d\n", length);
		break;

	case 'b' :
		if (!inet_aton(optarg, &bindaddr.sin_addr))
			printf("Invalid source host address\n");
		break;

	case 'B' :
		burstsize = atoi(optarg);
		break;
	case 'v':
		verbose++;
		break;

	case 'd':
		histogram = 1;
		break;

	case 'L':
		latencies = 1;
		break;

	case 'i' :
		inblocksize = atoi(optarg);
		if (inblocksize < 2 || inblocksize > 1000000)
			fatal("Invalid inblock size %d\n", inblocksize);
		break;

	case 'o' :
		outblocksize = atoi(optarg);
		if (outblocksize < 2 || outblocksize > 60000)
			fatal("Invalid outblocksize number %d\n", outblocksize);
		break;

	case 'r' :
		rate = atoi(optarg);
		if (rate > 1000000 || rate < 10)
			fatal("Invalid rate %d\n", rate);

		wait_time.tv_nsec = 1000000000 / rate;
		break;

	case 'j' :
		if (multicastgroups == MAX_MULTICASTGROUPS)
			fatal("Too many multicastgroups joined\n");

		if (!inet_aton(optarg, &multicastgroup[multicastgroups].sin_addr))
			fatal("Invalid Multicast address\n");
		else
			multicastgroups++;
		break;

	case 's':
		time_sync = 1;
		break;

	case 'X':
		terminate_remote = 1;
		break;

	case 'S':

		time_offset = atoi(optarg);
		break;

	case 'A':
		aggregate = 1;
		break;

	case 'W':
		warmup = 1;
		break;

	case 'C':
		if (strcasecmp(optarg, "A") ==0)
			class = 1;
		else if (strcasecmp(optarg, "B") == 0)
			class = 2;
		else if (strcasecmp(optarg, "C") == 0)
			class = 3;
		else {
			usage();
			return 4;
		}
		break;
	default:
		fatal("%s: Invalid option '%c'\n", argv[0], optopt);
	}

	if (argc > 1 + optind) {
		usage();
		return 1;
	}

	if (argc == 1 + optind) {
		if (!inet_aton(argv[optind], &addr.sin_addr))
			fatal("Invalid destination host address\n");

		if (warmup)
			return mcast_warmup(&addr);

		addr.sin_port = htons(port);
		bindaddr.sin_port = htons(origport);
		return mcast_send(&addr);
	} else {
		if (warmup) {
			usage();
			return 1;
		}
		bindaddr.sin_port = htons(port);
		return mcast_receive(&bindaddr);
	}
}

